#include "defs.h"
#include "menu.h"
#include "params.h"
#include "encoder.h"
#include "VdiLcd.h"
#include <LiquidCrystal_I2C.h>
#include <MenuSystem.h>

extern LiquidCrystal_I2C lcd;
extern CVDILcd vdi;
extern CParameters params;

// Menu variables
MenuSystem ms;
Menu mm("VDI menu:");


///////////////////////////////////////////////////////////////////////////////////////////////////////
MenuItem meShowVDI("Show VDI");
MenuItem meSetVDIStyle("VDI style");
MenuItem meSetMinSignal("Min sig val");
MenuItem meCalibZero("Calib ADC zero");
MenuItem meCalibMax("Calib ADC max");
MenuItem meShowSignal("Show raw signal");
MenuItem meShowSignalRange("Show sig range");
MenuItem meSetBacklight("Backlight");
MenuItem meLoadDefaults("Load defaults");
MenuItem meSetSound("Sound");
MenuItem meSetPolarity("Signal polarity");
MenuItem meSetOversampling("Averaging");


///////////////////////////////////////////////////////////////////////////////////////////////////////
//helper 
void onValueChangeHandler(const char* name, int* value, int minVal, int maxVal, int increment, int eepromIndex)
{
    lcd.clear();
    lcd.setCursor(0,0);
    lcd.print(name);
    lcd.setCursor(0,1);    
    lcd.print(*value);

    //wait until all activity stops
    bool prev(false);
    bool next(false);
    bool btnSelect(false);
    bool btnBack(false);   
    readEncoder(&prev, &next, &btnSelect, &btnBack);        
    while(btnSelect || btnBack)
    {
        delay(200);
        readEncoder(&prev, &next, &btnSelect, &btnBack);             
    }

    //change value until button click
    do
    {
        readEncoder(&prev, &next, &btnSelect, &btnBack);     
        if(prev)
            *value = *value - increment;
        if(next)
            *value = *value + increment;

        if(*value < minVal)
            *value = minVal;
        if(*value > maxVal)
            *value = maxVal;

        if(prev || next)
        {
            lcd.setCursor(0,1);
            lcd.print("                ");    
            lcd.setCursor(0,1);
            lcd.print(*value);
        }
        delay(1);               
    }while(!btnSelect && !btnBack);

    //remember
    EEPROMWriteInt(eepromIndex, *value);
    
    ms.back();           
    displayVDIMenu();      
}

void onShowVDISelected(MenuItem* p_menu_item)
{
    ms.back();           
    params.m_showVDI = true;    
    vdi.reset();
    vdi.updateVDI(params.m_vdiStyle, 0, 0, 0, false);
    delay(1000);    
}

void onSetVDIStyleSelected(MenuItem* p_menu_item)
{
    onValueChangeHandler("VDI style:", &params.m_vdiStyle, 0, 1, 1, 14);
}

void onSetMinSignalSelected(MenuItem* p_menu_item)
{
    onValueChangeHandler("Min signal val:", &params.m_minSignal, 0, 300, 1, 2);
}

void onCalibZeroSelected(MenuItem* p_menu_item)
{
    lcd.clear();
    lcd.setCursor(0,0);

    int X = analogRead(XPin) - params.m_XBias;
    int R = analogRead(RPin) - params.m_RBias;    

    lcd.print("Signals:");  
    lcd.setCursor(0,1); 
    lcd.print("X: ");  
    lcd.print(X);
    lcd.print(" R: "); 
    lcd.print(R); 
    delay(2000);

    lcd.clear();
    lcd.setCursor(0,0);
    lcd.print("Calibrating ......");

    const int samples = 1000;
    unsigned long lX = 0;
    unsigned long lR = 0;    
    for(int s=0; s<samples; s++)
    {
        delay(1);   
        lX = lX + analogRead(XPin);
        lR = lR + analogRead(RPin);  
    }
    params.m_XBias = lX/float(samples);
    params.m_RBias = lR/float(samples);    

    EEPROMWriteInt(4, params.m_XBias);
    EEPROMWriteInt(6, params.m_RBias);  

    lcd.clear();   
    lcd.setCursor(0,0);
    lcd.print("ADC zero:       ");  
    lcd.setCursor(0,1); 
    lcd.print("X: ");  
    lcd.print(params.m_XBias);
    lcd.print(" R: "); 
    lcd.print(params.m_RBias); 
    delay(3000);

    X = analogRead(XPin) - params.m_XBias;
    R = analogRead(RPin) - params.m_RBias;    

    lcd.clear();    
    lcd.setCursor(0,0);
    lcd.print("Signals:");  
    lcd.setCursor(0,1); 
    lcd.print("               ");  
    lcd.setCursor(0,1); 
    lcd.print("X: ");  
    lcd.print(X);
    lcd.print(" R: "); 
    lcd.print(R); 
    delay(2000);
    
    ms.back();           
    displayVDIMenu();      
}

void onCalibMaxSelected(MenuItem* p_menu_item)
{
    lcd.clear();
    lcd.setCursor(0,0);
    lcd.print("Current ADC max:");    
    lcd.setCursor(0,1);
    lcd.print(params.m_maxSignal);
    delay(2000);

    lcd.clear();
    lcd.setCursor(0,0);
    lcd.print("Calibrating ......");

    const int samples = 5000;
    int maxX = -1;
    int maxR = -1;    
    for(int s=0; s<samples; s++)
    {
        int X = analogRead(XPin);
        int R = analogRead(RPin);
        if(X>maxX) maxX = X;
        if(R>maxR) maxR = R;

        delay(1);         
    }

    int maxSignal = min(maxX, maxR) - 10;
    params.m_maxSignal = maxSignal;
    EEPROMWriteInt(8, params.m_maxSignal);  

    lcd.clear();   
    lcd.setCursor(0,0);
    lcd.print("ADC max signal:       ");  
    lcd.setCursor(0,1); 
    lcd.print(maxSignal);
    delay(3000);
    
    ms.back();           
    displayVDIMenu();      
}

void onShowSignalSelected(MenuItem* p_menu_item)
{
    bool prev(false);
    bool next(false);
    bool btnSelect(false);
    bool btnBack(false);   

    do
    {  
        #define SAMPLES 20
        int samples[SAMPLES];

        int pin[2] = {XPin, RPin};
        int avg[2];
        double stdDev[2];
        for(int i(0); i<2; i++)
        {
            double sum = 0;
            for(int n(0); n<SAMPLES;n++)
            {
              float adcV = 0.0F;
              for(int k(0);k<params.m_readingSamples;k++)
              {
                adcV += analogRead(pin[i]);
              }
              adcV /= params.m_readingSamples;

              samples[n] = adcV;
              sum += adcV;          

              delay(20);              
            }

            avg[i] = sum/SAMPLES;

            double variance = 0;
            for(int n(0); n<SAMPLES;n++)
            {
                variance += (samples[n] - avg[i]) * (samples[n] - avg[i]);
            }

            stdDev[i] = sqrt(variance/SAMPLES);
        }

        lcd.clear();
        lcd.setCursor(0, 0);
        lcd.print("X avg:");
        lcd.print(avg[0]);
        lcd.print(" \xE5:");
        lcd.print(int(stdDev[0]));

        lcd.setCursor(0, 1);
        lcd.print("R avg:");
        lcd.print(avg[1]);
        lcd.print(" \xE5:");
        lcd.print(int(stdDev[1]));
        delay(40);

        readEncoder(&prev, &next, &btnSelect, &btnBack);        

    }while(!btnSelect && !btnBack && !prev && !next);

    ms.back();           
    displayVDIMenu();      
}

void onShowSignalRangeSelected(MenuItem* p_menu_item)
{
    bool prev(false);
    bool next(false);
    bool btnSelect(false);
    bool btnBack(false);   

    do
    {  
        int minX = 0x5000;
        int minR = 0x5000;
        int maxX = 0;        
        int maxR = 0;        

        for(int n(0); n<500;n++)
        {
            int X = analogRead(XPin);
            if(X<minX) minX = X;
            if(X>maxX) maxX = X;

            int R = analogRead(RPin);
            if(R<minR) minR = R;
            if(R>maxR) maxR = R;

            delay(1);
        }

        lcd.clear();
        lcd.setCursor(0, 0);
        lcd.print("X: ");
        lcd.print(minX);
        lcd.print(" <-> ");
        lcd.print(maxX);

        lcd.setCursor(0, 1);
        lcd.print("R: ");
        lcd.print(minR);
        lcd.print(" <-> ");
        lcd.print(maxR);

        readEncoder(&prev, &next, &btnSelect, &btnBack);        
    }while(!btnSelect && !btnBack && !prev && !next);

    ms.back();           
    displayVDIMenu();      
}

void onSetBacklightSelected(MenuItem* p_menu_item)
{
    onValueChangeHandler("Backlight:", &params.m_LcdBacklight, 0, 1, 1, 12);
    if(params.m_LcdBacklight)  
        lcd.setBacklight(LED_ON);  
    else
        lcd.setBacklight(LED_OFF);      
    
    ms.back();           
    displayVDIMenu();      
}

void onLoadDefaultsSelected(MenuItem* p_menu_item)
{
    params.setDefaults();
    lcd.clear();
    lcd.setCursor(0,0);
    lcd.print("Values loaded");
    delay(1000);
    
    ms.back();           
    displayVDIMenu();      
}

void onSetSoundSelected(MenuItem* p_menu_item)
{
    onValueChangeHandler("Sound:", &params.m_soundOn, 0, 1, 1, 16);      
}

void onSetPolaritySelected(MenuItem* p_menu_item)
{
    onValueChangeHandler("Used R sig pol:", &params.m_sigPolarity, -1, 1, 1, 18);
}

void onSetOversampling(MenuItem* p_menu_item)
{
    onValueChangeHandler("Samples:", &params.m_readingSamples, 1, 30, 1, 10);
}

void createVDIMenu()
{
    mm.add_item(&meShowVDI, &onShowVDISelected);
    mm.add_item(&meSetVDIStyle, &onSetVDIStyleSelected);
    mm.add_item(&meSetMinSignal, &onSetMinSignalSelected);
    mm.add_item(&meShowSignal, &onShowSignalSelected);
    mm.add_item(&meShowSignalRange, &onShowSignalRangeSelected);
    mm.add_item(&meSetBacklight, onSetBacklightSelected);
    mm.add_item(&meSetOversampling, &onSetOversampling);
    mm.add_item(&meSetPolarity, &onSetPolaritySelected);
    mm.add_item(&meCalibZero, &onCalibZeroSelected);
    mm.add_item(&meCalibMax, &onCalibMaxSelected);
    mm.add_item(&meLoadDefaults, &onLoadDefaultsSelected);
    mm.add_item(&meSetSound, &onSetSoundSelected);

    ms.set_root_menu(&mm);
}

///////////////////////////////////////////////////////////////////////////////////////////////////////
//LCD menu; update this if you have different LCD (not 16x2)
void displayVDIMenu()
{
    if(params.m_showVDI)
      return;
    
    Menu const* cp_menu = ms.get_current_menu();

    lcd.clear();

    lcd.setCursor(0,0);
    lcd.print(cp_menu->get_name());

    lcd.setCursor(0,1);
    lcd.print(cp_menu->get_selected()->get_name());
}

